@@ -17,6 +17,7 @@ from typing import (
 )
 import io
 import os
+import pickle
 import sys
 import json
 import time
@@ -72,6 +73,8 @@ from .types import (
     DagList,
     DagPrettyNode,
     DagReload,
+    DagStatus,
+    DeleteBlobResponse,
     DynamicResults,
     DynamicStatusResponse,
     ESQueryResponse,
@@ -90,6 +93,7 @@ from .types import (
     ModelInfo,
     ModelParamsResponse,
     ModelReleaseResponse,
+    ModelVersionResponse,
     NamespaceList,
     NamespaceUpdateSettings,
     NodeChunk,
@@ -579,20 +583,30 @@ class XYMEClient:
 
     def get_dags(self) -> List[str]:
         return [
-            res[0]
+            res["dag"]
             for res in self.get_dag_times(retrieve_times=False)[1]
         ]
 
-    def get_dag_ages(self) -> List[Tuple[str, str, str]]:
+    def get_dag_ages(self) -> List[Dict[str, Optional[str]]]:
         cur_time, dags = self.get_dag_times(retrieve_times=True)
         return [
-            (dag_uri, get_age(cur_time, oldest), get_age(cur_time, latest))
-            for (dag_uri, oldest, latest) in sorted(dags, key=lambda el: (
-                safe_opt_num(el[1]), safe_opt_num(el[2]), el[0]))
+            {
+                "config_error": dag_status["config_error"],
+                "created": get_age(cur_time, dag_status["created"]),
+                "dag": dag_status["dag"],
+                "deleted": get_age(cur_time, dag_status["deleted"]),
+                "latest": get_age(cur_time, dag_status["latest"]),
+                "oldest": get_age(cur_time, dag_status["oldest"]),
+            }
+            for dag_status in sorted(dags, key=lambda el: (
+                el["config_error"] is None,
+                safe_opt_num(el["oldest"]),
+                safe_opt_num(el["latest"]),
+                el["dag"]))
         ]
 
-    def get_dag_times(self, retrieve_times: bool) -> Tuple[
-            float, List[Tuple[str, Optional[float], Optional[float]]]]:
+    def get_dag_times(self, retrieve_times: bool = True) -> Tuple[
+            float, List[DagStatus]]:
         res = cast(DagList, self.request_json(
             METHOD_GET, "/dags", {
                 "retrieve_times": int(retrieve_times),
@@ -722,10 +736,12 @@ class XYMEClient:
                 warnings_io.flush()
         return self.get_dag(dag_uri)
 
-    def set_settings(self, settings: SettingsObj) -> SettingsObj:
+    def set_settings(
+            self, config_token: str, settings: SettingsObj) -> SettingsObj:
         return cast(NamespaceUpdateSettings, self.request_json(
             METHOD_POST, "/settings", {
                 "settings": settings,
+                "config_token": config_token,
             }))["settings"]
 
     def get_settings(self) -> SettingsObj:
@@ -838,17 +854,24 @@ class XYMEClient:
             }))
 
     def get_named_secrets(
-            self, show_values: bool = False) -> Dict[str, Optional[str]]:
+            self,
+            config_token: Optional[str] = None,
+            show_values: bool = False) -> Dict[str, Optional[str]]:
+        if show_values and config_token is None:
+            raise ValueError("config_token must be set to show_values")
         return cast(Dict[str, Optional[str]], self.request_json(
             METHOD_GET, "/named_secrets", {
                 "show": int(bool(show_values)),
+                "config_token": config_token,
             }))
 
-    def set_named_secret(self, key: str, value: str) -> bool:
+    def set_named_secret(
+            self, config_token: str, key: str, value: str) -> bool:
         return cast(SetNamedSecret, self.request_json(
             METHOD_PUT, "/named_secrets", {
                 "key": key,
                 "value": value,
+                "config_token": config_token,
             }))["replaced"]
 
     def get_error_logs(self) -> str:
@@ -944,6 +967,13 @@ class XYMEClient:
         return cast(UUIDResponse, self.request_json(
             METHOD_GET, "/uuid", {}))["uuid"]
 
+    def delete_blobs(self, blob_uris: List[str]) -> DeleteBlobResponse:
+        return cast(DeleteBlobResponse, self.request_json(
+            METHOD_DELETE, "/blob", {
+                "blob_uris": blob_uris,
+            },
+        ))
+
 # *** XYMEClient ***
 
 
@@ -1346,20 +1376,25 @@ class DagHandle:
             self,
             nodes_only: bool,
             allow_unicode: bool,
-            method: Optional[str] = "accern") -> PrettyResponse:
+            method: Optional[str] = "accern",
+            fields: Optional[List[str]] = None) -> PrettyResponse:
+        args = {
+            "dag": self.get_uri(),
+            "nodes_only": nodes_only,
+            "allow_unicode": allow_unicode,
+            "method": method,
+        }
+        if fields is not None:
+            args["fields"] = ",".join(fields)
         return cast(PrettyResponse, self._client.request_json(
-            METHOD_GET, "/pretty", {
-                "dag": self.get_uri(),
-                "nodes_only": nodes_only,
-                "allow_unicode": allow_unicode,
-                "method": method,
-            }))
+            METHOD_GET, "/pretty", args))
 
     def pretty(
             self,
             nodes_only: bool = False,
             allow_unicode: bool = True,
             method: Optional[str] = "dot",
+            fields: Optional[List[str]] = None,
             output_format: Optional[str] = "png",
             display: Optional[IO[Any]] = sys.stdout) -> Optional[str]:
 
@@ -1373,7 +1408,8 @@ class DagHandle:
         graph_str = self._pretty(
             nodes_only=nodes_only,
             allow_unicode=allow_unicode,
-            method=method)["pretty"]
+            method=method,
+            fields=fields)["pretty"]
         if method == "accern":
             return render(graph_str)
         if method == "dot":
@@ -1437,9 +1473,12 @@ class DagHandle:
     def pretty_obj(
             self,
             nodes_only: bool = False,
-            allow_unicode: bool = True) -> List[DagPrettyNode]:
+            allow_unicode: bool = True,
+            fields: Optional[List[str]] = None) -> List[DagPrettyNode]:
         return self._pretty(
-            nodes_only=nodes_only, allow_unicode=allow_unicode)["nodes"]
+            nodes_only=nodes_only,
+            allow_unicode=allow_unicode,
+            fields=fields)["nodes"]
 
     def get_def(self, full: bool = True) -> DagDef:
         return cast(DagDef, self._client.request_json(
@@ -1488,13 +1527,13 @@ class DagHandle:
             MinimalQueueStatsResponse, QueueStatsResponse]:
         return self._client.check_queue_stats(self.get_uri(), minimal=minimal)
 
-    def scale_worker(self, replicas: int) -> bool:
+    def scale_worker(self, replicas: int) -> int:
         return cast(WorkerScale, self._client.request_json(
             METHOD_PUT, "/worker", {
                 "dag": self.get_uri(),
                 "replicas": replicas,
                 "task": None,
-            }))["success"]
+            }))["num_replicas"]
 
     def reload(self, timestamp: Optional[float] = None) -> float:
         return cast(DagReload, self._client.request_json(
@@ -1698,6 +1737,13 @@ class DagHandle:
                 **kwargs,
             }))
 
+    def delete(self) -> DeleteBlobResponse:
+        return cast(DeleteBlobResponse, self._client.request_json(
+            METHOD_DELETE, "/blob", {
+                "blob_uris": [self.get_uri()],
+            },
+        ))
+
     def __hash__(self) -> int:
         return hash(self.get_uri())
 
@@ -2325,6 +2371,31 @@ class BlobHandle:
             METHOD_POST, "/finish_zip", {"uri": uri}))
         return res["files"]
 
+    def _finish_upload_sklike(
+            self,
+            xcols: List[str],
+            is_clf: bool,
+            model_name: str,
+            maybe_classes: Optional[List[str]],
+            maybe_range: Tuple[Optional[float], Optional[float]],
+            full_init: bool) -> UploadFilesResponse:
+        uri = self._tmp_uri
+        if uri is None:
+            raise ValueError("tmp_uri is None")
+        return cast(UploadFilesResponse, self._client.request_json(
+            METHOD_POST, "/finish_sklike", {
+                "classes": maybe_classes,
+                "full_init": full_init,
+                "is_clf": is_clf,
+                "model_uri": self.get_uri(),
+                "model_name": model_name,
+                "output_range": maybe_range,
+                "owner_dag": self.get_owner_dag(),
+                "owner_node": self.get_owner_node(),
+                "tmp_uri": uri,
+                "xcols": xcols,
+            }))
+
     def _clear_upload(self) -> None:
         uri = self._tmp_uri
         if uri is None:
@@ -2375,6 +2446,58 @@ class BlobHandle:
             for blob_uri in files
         ]
 
+    def upload_sklike_model_file(
+            self,
+            model_obj: IO[bytes],
+            xcols: List[str],
+            is_clf: bool,
+            model_name: str,
+            maybe_classes: Optional[List[str]] = None,
+            maybe_range: Optional[
+                Tuple[Optional[float], Optional[float]]] = None,
+            full_init: bool = True) -> UploadFilesResponse:
+        try:
+            self._upload_file(model_obj, ext="pkl")
+            output_range = (None, None) if maybe_range is None else maybe_range
+            return self._finish_upload_sklike(
+                model_name=model_name,
+                maybe_classes=maybe_classes,
+                maybe_range=output_range,
+                xcols=xcols,
+                is_clf=is_clf,
+                full_init=full_init)
+        finally:
+            self._clear_upload()
+
+    def upload_sklike_model(
+            self,
+            model: Any,
+            xcols: List[str],
+            is_clf: bool,
+            maybe_classes: Optional[List[str]] = None,
+            maybe_range: Optional[
+                Tuple[Optional[float], Optional[float]]] = None,
+            full_init: bool = True) -> UploadFilesResponse:
+        try:
+            model_name = type(model).__name__
+        except Exception as e:
+            raise ValueError(f"can not infer model name {model}") from e
+        try:
+            if is_clf and maybe_classes is None:
+                maybe_classes = model.classes_
+        except Exception as e:
+            raise ValueError(f"can not infer classes from {model}") from e
+        dump = pickle.dumps(model, pickle.HIGHEST_PROTOCOL)
+        with io.BytesIO(dump) as buffer:
+            return self.upload_sklike_model_file(
+                buffer,
+                xcols,
+                is_clf,
+                model_name,
+                maybe_classes,
+                maybe_range,
+                full_init)
+
     def convert_model(self, reload: bool = True) -> ModelReleaseResponse:
         return cast(ModelReleaseResponse, self._client.request_json(
             METHOD_POST, "/convert_model", {
@@ -2382,12 +2505,57 @@ class BlobHandle:
                 "reload": reload,
             }))
 
+    def delete(self) -> DeleteBlobResponse:
+        return cast(DeleteBlobResponse, self._client.request_json(
+            METHOD_DELETE, "/blob", {
+                "blob_uris": [self.get_uri()],
+            },
+        ))
+
     def get_model_release(self) -> ModelReleaseResponse:
         return cast(ModelReleaseResponse, self._client.request_json(
             METHOD_GET, "/model_release", {
                 "blob": self.get_uri(),
             }))
 
+    def get_model_version(self) -> ModelVersionResponse:
+        return cast(ModelVersionResponse, self._client.request_json(
+            METHOD_GET, "/model_version", {
+                "model_uri": self.get_uri(),
+            }))
+
+    def _copy_model_version(
+            self,
+            model_uri: str,
+            read_version: Optional[int],
+            write_version: int,
+            overwrite: bool) -> ModelVersionResponse:
+        return cast(ModelVersionResponse, self._client.request_json(
+            METHOD_PUT, "/model_version", {
+                "model_uri": model_uri,
+                "read_version": read_version,
+                "write_version": write_version,
+                "overwrite": overwrite,
+            }))
+
+    def copy_model_version(
+            self,
+            read_version: int,
+            write_version: int,
+            overwrite: bool) -> ModelVersionResponse:
+        return self._copy_model_version(
+            model_uri=self.get_uri(),
+            read_version=read_version,
+            write_version=write_version,
+            overwrite=overwrite)
+
+    def delete_model_version(self, version: int) -> ModelVersionResponse:
+        return self._copy_model_version(
+            model_uri=self.get_uri(),
+            read_version=None,
+            write_version=version,
+            overwrite=True)
+
     def __hash__(self) -> int:
         return hash(self.as_str())
 
@@ -2412,13 +2580,13 @@ class CSVBlobHandle(BlobHandle):
     def finish_csv_upload(
             self, filename: Optional[str] = None) -> UploadFilesResponse:
         tmp_uri = self._tmp_uri
-        assert tmp_uri is not None
-        owner = self.get_owner()
+        if tmp_uri is None:
+            raise ValueError("tmp_uri is None")
         args: Dict[str, Optional[Union[str, int]]] = {
             "tmp_uri": tmp_uri,
             "csv_uri": self.get_uri(),
-            "owner_dag": owner["owner_dag"],
-            "owner_node": owner["owner_node"],
+            "owner_dag": self.get_owner_dag(),
+            "owner_node": self.get_owner_node(),
             "filename": filename,
         }
         return cast(UploadFilesResponse, self._client.request_json(
@@ -2540,7 +2708,7 @@ class CustomCodeBlobHandle(BlobHandle):
         def fn_as_str(fun: FUNC) -> str:
             body = textwrap.dedent(inspect.getsource(fun))
             res = body + textwrap.dedent(f"""
-            result = {fun.__name__}(*data)
+            result = {fun.__name__}(*data, **kwargs)
             if result is None:
                 raise ValueError("{fun.__name__} must return a value")
             """)
