@@ -170,9 +170,17 @@ NodeInfo = TypedDict('NodeInfo', {
     "state": Optional[int],
     "config_error": Optional[str],
 })
+DagStatus = TypedDict("DagStatus", {
+    "config_error": Optional[str],
+    "created": Optional[float],
+    "dag": str,
+    "deleted": Optional[float],
+    "latest": Optional[float],
+    "oldest": Optional[float],
+})
 DagList = TypedDict('DagList', {
     "cur_time": float,
-    "dags": List[Tuple[str, Optional[float], Optional[float]]],
+    "dags": List[DagStatus],
 })
 KnownBlobs = TypedDict('KnownBlobs', {
     "cur_time": float,
@@ -268,7 +276,7 @@ FlushAllQueuesResponse = TypedDict('FlushAllQueuesResponse', {
     "success": bool,
 })
 WorkerScale = TypedDict('WorkerScale', {
-    "success": bool,
+    "num_replicas": int,
 })
 SetNamedSecret = TypedDict('SetNamedSecret', {
     "replaced": bool,
@@ -357,9 +365,17 @@ QueueMode = TypedDict('QueueMode', {
 ModelReleaseResponse = TypedDict('ModelReleaseResponse', {
     "release": Optional[int],
 })
+ModelVersionResponse = TypedDict('ModelVersionResponse', {
+    "max_version": Optional[int],
+    "all_versions": List[str],
+})
 ESQueryResponse = TypedDict('ESQueryResponse', {
     "query": Dict[str, Any],
 })
+
+PrettyField = Literal[
+    "chunk",
+]
 DagPrettyEdge = TypedDict('DagPrettyEdge', {
     "out_name": str,
     "in_node": str,
@@ -372,7 +388,7 @@ DagPrettyNode = TypedDict('DagPrettyNode', {
     "id": str,
     "name": str,
     "kind": str,
-    "chunk": int,
+    "info": Dict[PrettyField, Union[str, int]],
     "high_ixs": Dict[str, int],
     "status": 'TaskStatus',
     "out_edges": List[DagPrettyEdge],
@@ -382,6 +398,7 @@ PrettyResponse = TypedDict('PrettyResponse', {
     "pretty": str,
     "nodes": List[DagPrettyNode],
 })
+
 BlobURIResponse = TypedDict('BlobURIResponse', {
     "uri": str,
     "owner": BlobOwner,
@@ -398,3 +415,11 @@ UUIDResponse = TypedDict('UUIDResponse', {
 NodeTypeResponse = TypedDict('NodeTypeResponse', {
     "is_model": bool,
 })
+DeleteBlobStatus = TypedDict('DeleteBlobStatus', {
+    "uri": str,
+    "status": str,
+    "deletion_time": Optional[float],
+})
+DeleteBlobResponse = TypedDict('DeleteBlobResponse', {
+    "blobs": List[DeleteBlobStatus],
+})
